import os
from pathlib import Path
import shutil

import jinja2
import toml

from md2site.post import Post, name_to_slug
from md2site.site import PostMetaData, Site

from md2site.renderer import Parser, Renderer, extract_wikilinks


def prepare_output_folder():
    output_folder = Path("dist")
    if output_folder.exists():
        shutil.rmtree(output_folder, ignore_errors=True)
    os.makedirs(output_folder)


def load_config() -> Site:
    raw_config = toml.load("config.toml")
    name = raw_config["name"]
    base_url = raw_config["base_url"]
    return Site(base_url, name)


def parse_markdown(markdown: str) -> str:
    return markdown


def load_post_files() -> list[Post]:
    posts = list()
    for child in Path("posts").iterdir():
        if child.name.endswith(".md"):
            posts.append(Post.from_file(child))
    posts.sort(key=lambda p: p.created_at, reverse=True)
    return posts


def write_html(site: Site, post: Post, template: jinja2.Template):
    with open(f"dist/{post.slug}.html", "w", encoding="utf-8") as file:
        html = template.render(post=post, site=site)
        file.write(html)


def build_posts(posts: list[Post], site: Site):
    loader = jinja2.FileSystemLoader("template")
    env = jinja2.Environment(loader=loader)
    post_template = env.get_template("post.html")
    index_template = env.get_template("index.html")
    renderer = Renderer(site)
    parser = Parser()
    for post in posts:
        parsed = parser.parse(post.content)
        post.content = renderer.render(parsed)
        if post.name == "index":
            write_html(site, post, index_template)
        else:
            write_html(site, post, post_template)


def populate_backlinks(posts: list[Post], base_url: str):
    backlink_map = build_backlink_map(posts)
    for post in posts:
        if post.name not in backlink_map:
            continue
        backlinks = []
        for name in backlink_map[post.name]:
            backlinks.append({"title": name, "url": name_to_url(name, base_url)})
        post.backlinks = backlinks


def name_to_url(name: str, base_url: str) -> str:
    return f"{base_url}/{name_to_slug(name)}.html"


def build_backlink_map(posts: list[Post]) -> dict[str, set[str]]:
    result = {}
    for p in posts:
        links = extract_wikilinks(p.content)
        for link in links:
            if link not in result:
                result[link] = set()
            result[link].add(p.name)
    return result


def build_link_map(posts: list[Post]) -> dict[str, str]:
    mapping = {}
    for post in posts:
        mapping[post.name] = post.slug
    return mapping


def copy_static_files():
    """
    copy every non-html file inside template/ folder to dist/.
    """
    output_folder = Path("./dist")
    for file in Path("./template").iterdir():
        if file.name.endswith(".html"):
            continue
        shutil.copyfile(file, output_folder / file.name)


def generate():
    prepare_output_folder()
    copy_static_files()
    site = load_config()
    posts = load_post_files()
    site.link_map = build_link_map(posts)
    site.recent_posts = [
        PostMetaData(p.name, p.title, p.created_at, name_to_url(p.name, site.base_url))
        for p in posts[:10]
    ]
    populate_backlinks(posts, site.base_url)
    build_posts(posts, site)



import re
from mistletoe import Document
from mistletoe import span_token

from mistletoe.html_renderer import HTMLRenderer
from mistletoe.span_token import RawText, SpanToken

from md2site.site import Site
from mistletoe.token import Token


class WikiLink(SpanToken):
    """
    Obsidian-style wiki link token.
    `[[page_name]]` or `[[page_name|link_text]]`.

    This is an inline token with a single child of type RawText.
    """

    pattern = re.compile(r"\[\[([\s\S]+?[\s\S]+?)\]\]")
    parse_inner = False

    def __init__(self, match: re.Match):
        super().__init__(match)
        self.target: str = ""
        link_text = match.group(1)
        splits = link_text.split("|", 1)
        self.target = splits[0].strip()
        if len(splits) == 1:
            self.children = (RawText(self.target),)
        else:
            self.children = (RawText(splits[1].strip()),)


class Renderer(HTMLRenderer):
    def __init__(self, config: Site):
        super().__init__(WikiLink)
        self.site = config

    def render_wiki_link(self, token: WikiLink):
        template = '<a href="{base_url}/{target}.html">{inner}</a>'
        target = self.site.link_map.get(token.target, "#")
        inner = self.render_inner(token)
        return template.format(base_url=self.site.base_url, target=target, inner=inner)


class Parser:
    def __init__(self):
        span_token.add_token(WikiLink)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_val, traceback):
        span_token.reset_tokens()

    def parse(self, markdown: str):
        document = Document(markdown)
        return document


def extract_wikilinks(markdown: str) -> set[str]:
    with Parser() as p:
        ast = p.parse(markdown)
        links = set()
        nodes: list[Token] = [ast]
        while nodes:
            node = nodes.pop(0)
            if isinstance(node, WikiLink):
                links.add(node.target)
            if isinstance(node, RawText):
                continue
            if hasattr(node, "children"):
                for child in node.children:
                    nodes.append(child)
        return links



from dataclasses import dataclass, field
from datetime import datetime


@dataclass
class PostMetaData:
    name: str
    title: str
    created_at: datetime
    url: str


@dataclass
class Site:
    """
    Site configuration loaded from config.toml.
    """

    base_url: str
    name: str
    link_map: dict[str, str] = field(default_factory=dict)
    recent_posts: list[PostMetaData] = field(default_factory=list)



from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import TextIO
import os

import yaml


@dataclass
class MarkdownContent:
    frontmatter: dict[str, str]
    body: str


@dataclass
class Post:
    """
    All information necessary to render a post.
    """

    name: str
    title: str
    content: str
    summary: str
    created_at: datetime
    slug: str
    backlinks: list[str] = field(default_factory=list)

    @classmethod
    def from_file(cls, filepath: Path) -> Post:
        with open(filepath, "r", encoding="utf-8") as f:
            content = separate_frontmatter(f)
            name = filepath.stem
            if "title" in content.frontmatter:
                title = content.frontmatter["title"]
            else:
                title = name

            if "date" in content.frontmatter:
                created_at = content.frontmatter["date"]
            else:
                created_at = datetime.fromtimestamp(os.path.getctime(filepath)).replace(
                    tzinfo=timezone.utc
                )
            return Post(
                name=name,
                title=title,
                content=content.body,
                summary=content.body[:140],
                created_at=created_at,
                slug=name_to_slug(name),
            )


def parse_frontmatter(yaml_text: str) -> dict[str, str]:
    parsed = yaml.safe_load(yaml_text)
    if parsed is not None:
        return parsed
    else:
        return {}


def separate_frontmatter(markdown: TextIO) -> MarkdownContent:
    is_frontmatter = False
    frontmatter_lines: list[str] = []
    body_lines: list[str] = []
    for line_number, line in enumerate(markdown):
        if line_number == 0 and line.strip() == "---":
            is_frontmatter = True
            continue
        if is_frontmatter and line.strip() == "---":
            is_frontmatter = False
            continue
        if is_frontmatter:
            frontmatter_lines.append(line)
        else:
            body_lines.append(line)

    frontmatter = parse_frontmatter("".join(frontmatter_lines))
    body = "".join(body_lines)
    return MarkdownContent(frontmatter, body)


def name_to_slug(name: str) -> str:
    return name.lower().replace(" ", "-")
